@@ -1,25 +1,27 @@
-from __future__ import annotations
-
-import dataclasses
-import warnings
-
 """
 Flask-Limiter Extension
 """
+from __future__ import annotations
+
+import dataclasses
 import datetime
 import itertools
 import logging
 import time
+import traceback
+import warnings
 import weakref
 from collections import defaultdict
 from functools import partial, wraps
-from typing import overload
+from types import TracebackType
+from typing import Type, overload
 
 import flask
 import flask.wrappers
 from limits.errors import ConfigurationError
 from limits.storage import MemoryStorage, Storage, storage_from_string
 from limits.strategies import STRATEGIES, RateLimiter
+from ordered_set import OrderedSet
 from werkzeug.http import http_date, parse_date
 
 from ._compat import request_context
@@ -40,32 +42,33 @@ from .typing import (
     Union,
     cast,
 )
+from .util import get_qualified_name
 from .wrappers import Limit, LimitGroup, RequestLimit
 
 
 @dataclasses.dataclass
 class LimiterContext:
-    rate_limiting_complete: bool = False
     view_rate_limit: Optional[RequestLimit] = None
     view_rate_limits: List[RequestLimit] = dataclasses.field(default_factory=list)
     conditional_deductions: Dict[Limit, List[str]] = dataclasses.field(
         default_factory=dict
     )
+    seen_limits: OrderedSet[Limit] = dataclasses.field(default_factory=OrderedSet)
 
     def reset(self) -> None:
-        self.rate_limiting_complete = False
         self.view_rate_limit = None
         self.view_rate_limits.clear()
         self.conditional_deductions.clear()
+        self.seen_limits.clear()
 
 
 class Limiter:
     """
     The :class:`Limiter` class initializes the Flask-Limiter extension.
 
-    :param app: :class:`flask.Flask` instance to initialize the extension with.
     :param key_func: a callable that returns the domain to rate limit
       by.
+    :param app: :class:`flask.Flask` instance to initialize the extension with.
     :param default_limits: a variable list of strings or callables
      returning strings denoting default limits to apply to all routes.
      :ref:`ratelimit-string` for  more details.
@@ -97,7 +100,7 @@ class Limiter:
      upon instantiation.
     :param auto_check: whether to automatically check the rate limit in
      the before_request chain of the application. default ``True``
-    :param swallow_errors: whether to swallow errors when hitting a rate
+    :param swallow_errors: whether to swallow any errors when hitting a rate
      limit. An exception will still be logged. default ``False``
     :param fail_on_first_breach: whether to stop processing remaining limits
      after the first breach. default ``True``
@@ -114,13 +117,16 @@ class Limiter:
     :param retry_after: Allows configuration of how the value of the
      `Retry-After` header is rendered. One of `http-date` or `delta-seconds`.
     :param key_prefix: prefix prepended to rate limiter keys and app context global names.
+    :param request_identifier: a callable that returns the unique identity the current request.
+     Defaults to :attr:`flask.Request.endpoint`
     :param enabled: Whether the extension is enabled or not
     """
 
     def __init__(
         self,
+        key_func: Callable[[], str],
+        *,
         app: Optional[flask.Flask] = None,
-        key_func: Optional[Callable[[], str]] = None,
         default_limits: Optional[List[Union[str, Callable[[], str]]]] = None,
         default_limits_per_method: Optional[bool] = None,
         default_limits_exempt_when: Optional[Callable[[], bool]] = None,
@@ -145,6 +151,7 @@ class Limiter:
         in_memory_fallback_enabled: Optional[bool] = None,
         retry_after: Optional[str] = None,
         key_prefix: str = "",
+        request_identifier: Optional[Callable[..., str]] = None,
         enabled: bool = True,
     ) -> None:
         self.app = app
@@ -180,26 +187,15 @@ class Limiter:
         self._fail_on_first_breach = fail_on_first_breach
         self._on_breach = on_breach
 
-        # No longer optional
-        assert key_func
-
         self._key_func = key_func
         self._key_prefix = key_prefix
+        self._request_identifier = request_identifier
 
         _default_limits = (
             [
                 LimitGroup(
-                    limit,
-                    self._key_func,
-                    None,
-                    False,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    1,
+                    limit_provider=limit,
+                    key_function=self._key_func,
                 )
                 for limit in default_limits
             ]
@@ -210,17 +206,10 @@ class Limiter:
         _application_limits = (
             [
                 LimitGroup(
-                    limit,
-                    self._key_func,
-                    "global",
-                    False,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    1,
+                    limit_provider=limit,
+                    key_function=self._key_func,
+                    scope="global",
+                    shared=True,
                 )
                 for limit in application_limits
             ]
@@ -232,17 +221,8 @@ class Limiter:
             for limit in in_memory_fallback:
                 self._in_memory_fallback.append(
                     LimitGroup(
-                        limit,
-                        self._key_func,
-                        None,
-                        False,
-                        None,
-                        None,
-                        None,
-                        None,
-                        None,
-                        None,
-                        1,
+                        limit_provider=limit,
+                        key_function=self._key_func,
                     )
                 )
 
@@ -260,10 +240,8 @@ class Limiter:
         self.limit_manager = LimitManager(
             application_limits=_application_limits,
             default_limits=_default_limits,
-            static_route_limits={},
-            dynamic_route_limits={},
-            static_blueprint_limits={},
-            dynamic_blueprint_limits={},
+            decorated_limits={},
+            blueprint_limits={},
             route_exemptions=self._route_exemptions,
             blueprint_exemptions=self._blueprint_exemptions,
         )
@@ -309,9 +287,7 @@ class Limiter:
             self._headers_enabled = bool(config.get(ConfigVars.HEADERS_ENABLED, False))
 
         self._storage_options.update(config.get(ConfigVars.STORAGE_OPTIONS, {}))
-        storage_uri_from_config = config.get(
-            ConfigVars.STORAGE_URI, config.get(ConfigVars.STORAGE_URL, None)
-        )
+        storage_uri_from_config = config.get(ConfigVars.STORAGE_URI, None)
         if not storage_uri_from_config:
             if not self._storage_uri:
                 warnings.warn(
@@ -362,7 +338,9 @@ class Limiter:
         )
 
         self._key_prefix = self._key_prefix or config.get(ConfigVars.KEY_PREFIX, "")
-
+        self._request_identifier = self._request_identifier or config.get(
+            ConfigVars.REQUEST_IDENTIFIER, lambda: flask.request.endpoint or ""
+        )
         app_limits = config.get(ConfigVars.APPLICATION_LIMITS, None)
         self._application_limits_cost = self._application_limits_cost or config.get(
             ConfigVars.APPLICATION_LIMITS_COST, 1
@@ -372,17 +350,11 @@ class Limiter:
             self.limit_manager.set_application_limits(
                 [
                     LimitGroup(
-                        app_limits,
-                        self._key_func,
-                        "global",
-                        False,
-                        None,
-                        None,
-                        None,
-                        None,
-                        None,
-                        None,
-                        self._application_limits_cost,
+                        limit_provider=app_limits,
+                        key_function=self._key_func,
+                        scope="global",
+                        shared=True,
+                        cost=self._application_limits_cost,
                     )
                 ]
             )
@@ -398,17 +370,12 @@ class Limiter:
             self.limit_manager.set_default_limits(
                 [
                     LimitGroup(
-                        conf_limits,
-                        self._key_func,
-                        None,
-                        self._default_limits_per_method,
-                        None,
-                        None,
-                        self._default_limits_exempt_when,
-                        None,
-                        self._default_limits_deduct_when,
-                        None,
-                        self._default_limits_cost,
+                        limit_provider=conf_limits,
+                        key_function=self._key_func,
+                        per_method=self._default_limits_per_method,
+                        exempt_when=self._default_limits_exempt_when,
+                        deduct_when=self._default_limits_deduct_when,
+                        cost=self._default_limits_cost,
                     )
                 ]
             )
@@ -422,10 +389,6 @@ class Limiter:
             self.limit_manager.set_default_limits(default_limit_groups)
         self.__configure_fallbacks(app, self._strategy)
 
-        # purely for backward compatibility as stated in flask documentation
-        if not hasattr(app, "extensions"):
-            app.extensions = {}  # pragma: no cover
-
         if self not in app.extensions.setdefault("limiter", set()):
             if self._auto_check:
                 app.before_request(self._check_request_limit)
@@ -437,17 +400,27 @@ class Limiter:
 
     @property
     def context(self) -> LimiterContext:
+        """
+        The context is meant to exist for the lifetime
+        of a request/response cycle per instance of the extension
+        so as to keep track of any state used at different steps
+        in the lifecycle (for example to pass information
+        from the before request hook to the after_request hook)
+
+        :meta private:
+        """
         ctx = request_context()
         if not hasattr(ctx, "_limiter_request_context"):
             ctx._limiter_request_context = defaultdict(LimiterContext)  # type: ignore
         return cast(
-            Dict[str, LimiterContext],
+            Dict[Limiter, LimiterContext],
             ctx._limiter_request_context,  # type: ignore
-        )[self._key_prefix]
+        )[self]
 
     def limit(
         self,
         limit_value: Union[str, Callable[[], str]],
+        *,
         key_func: Optional[Callable[[], str]] = None,
         per_method: bool = False,
         methods: Optional[List[str]] = None,
@@ -459,9 +432,10 @@ class Limiter:
             Callable[[RequestLimit], Optional[flask.wrappers.Response]]
         ] = None,
         cost: Union[int, Callable[[], int]] = 1,
+        scope: Optional[Union[str, Callable[[str], str]]] = None,
     ) -> LimitDecorator:
         """
-        decorator to be used for rate limiting individual routes or blueprints.
+        Decorator to be used for rate limiting individual routes or blueprints.
 
         :param limit_value: rate limit string or a callable that returns a
          string. :ref:`ratelimit-string` for more details.
@@ -492,12 +466,29 @@ class Limiter:
          raised.
         :param cost: The cost of a hit or a function that
          takes no parameters and returns the cost as an integer (Default: ``1``).
+        :param scope: a string or callable that returns a string
+         for further categorizing the rate limiting scope. This scope is combined
+         with the current endpoint of the request.
+
+
+        Changes
+          - .. versionadded:: 2.9.0 The returned object can also be used as a context manager
+            for rate limiting a code block inside a view. For example::
+
+                @app.route("/")
+                def route():
+                   try:
+                       with limiter.limit("10/second"):
+                           # something expensive
+                   except RateLimitExceeded: pass
         """
 
         return LimitDecorator(
             self,
             limit_value,
             key_func,
+            False,
+            scope,
             per_method=per_method,
             methods=methods,
             error_message=error_message,
@@ -512,7 +503,10 @@ class Limiter:
         self,
         limit_value: Union[str, Callable[[], str]],
         scope: Union[str, Callable[[str], str]],
+        *,
         key_func: Optional[Callable[[], str]] = None,
+        per_method: bool = False,
+        methods: Optional[List[str]] = None,
         error_message: Optional[str] = None,
         exempt_when: Optional[Callable[[], bool]] = None,
         override_defaults: bool = True,
@@ -532,6 +526,10 @@ class Limiter:
         :param key_func: function/lambda to extract the unique
          identifier for the rate limit. defaults to remote address of the
          request.
+        :param per_method: whether the limit is sub categorized into the
+         http method of the request.
+        :param methods: if specified, only the methods in this list will
+         be rate limited (default: ``None``).
         :param error_message: string (or callable that returns one) to override
          the error message used in the response.
         :param function exempt_when: function/lambda used to decide if the rate
@@ -559,6 +557,8 @@ class Limiter:
             key_func,
             True,
             scope,
+            per_method=per_method,
+            methods=methods,
             error_message=error_message,
             exempt_when=exempt_when,
             override_defaults=override_defaults,
@@ -648,9 +648,7 @@ class Limiter:
         if isinstance(obj, flask.Blueprint):
             self.limit_manager.add_blueprint_exemption(obj.name, flags)
         elif obj:
-            self.limit_manager.add_route_exemption(
-                f"{obj.__module__}.{obj.__name__}", flags
-            )
+            self.limit_manager.add_route_exemption(get_qualified_name(obj), flags)
         else:
             _R = TypeVar("_R")
             _WO = TypeVar("_WO", Callable[..., _R], flask.Blueprint)
@@ -681,17 +679,11 @@ class Limiter:
         if not self._in_memory_fallback and fallback_limits:
             self._in_memory_fallback = [
                 LimitGroup(
-                    fallback_limits,
-                    self._key_func,
-                    None,
-                    False,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    1,
+                    limit_provider=fallback_limits,
+                    key_function=self._key_func,
+                    scope=None,
+                    per_method=False,
+                    cost=1,
                 )
             ]
 
@@ -725,7 +717,7 @@ class Limiter:
 
         :raises: RateLimitExceeded
         """
-        self._check_request_limit(False)
+        self._check_request_limit(in_middleware=False)
 
     def reset(self) -> None:
         """
@@ -796,11 +788,31 @@ class Limiter:
 
         return self.context.view_rate_limits
 
-    def __check_conditional_deductions(self, response: flask.wrappers.Response) -> None:
+    def identify_request(self) -> str:
+        """
+        Returns the identity of the request (by default this is the
+        :attr:`flask.Request.endpoint` associated by the view function
+        that is handling the request). The behavior can be customized
+        by initializing the extension with a callable argument for
+        :paramref:`~flask_limiter.Limiter.request_identifier`.
+        """
+        if self.initialized and self.enabled:
+            assert self._request_identifier
+            return self._request_identifier()
+        return ""
 
+    def __check_conditional_deductions(self, response: flask.wrappers.Response) -> None:
         for lim, args in self.context.conditional_deductions.items():
             if lim.deduct_when and lim.deduct_when(response):
-                self.limiter.hit(lim.limit, *args, cost=lim.cost)
+                try:
+                    self.limiter.hit(lim.limit, *args, cost=lim.cost)
+                except Exception as err:
+                    if self._swallow_errors:
+                        self.logger.exception(
+                            "Failed to deduct rate limit. " "Swallowing error"
+                        )
+                    else:
+                        raise err
 
     def __inject_headers(
         self, response: flask.wrappers.Response
@@ -868,28 +880,37 @@ class Limiter:
 
         return response
 
-    def __check_all_limits_exempt(self, endpoint: Optional[str]) -> bool:
+    def __check_all_limits_exempt(
+        self,
+        endpoint: Optional[str],
+    ) -> bool:
         return bool(
             not endpoint
             or not (self.enabled and self.initialized)
             or endpoint == "static"
             or any(fn() for fn in self._request_filters)
-            or self.context.rate_limiting_complete
         )
 
     def __filter_limits(
         self,
         endpoint: Optional[str],
         blueprint: Optional[str],
+        callable_name: Optional[str],
         in_middleware: bool = False,
     ) -> List[Limit]:
-        view_func = flask.current_app.view_functions.get(endpoint or "", None)
-        name = f"{view_func.__module__}.{view_func.__name__}" if view_func else ""
+        if callable_name:
+            name = callable_name
+        else:
+            view_func = flask.current_app.view_functions.get(endpoint or "", None)
+            name = get_qualified_name(view_func) if view_func else ""
 
         if self.__check_all_limits_exempt(endpoint):
             return []
 
-        marked_for_limiting = name in self._marked_for_limiting
+        marked_for_limiting = (
+            name in self._marked_for_limiting
+            or self.limit_manager.has_hints(endpoint or "")
+        )
         fallback_limits = []
 
         if self._storage_dead and self._fallback_limiter:
@@ -906,28 +927,30 @@ class Limiter:
                     self.__check_backend_count = 0
                 else:
                     fallback_limits = list(itertools.chain(*self._in_memory_fallback))
+        if fallback_limits:
+            return fallback_limits
 
-        return self.limit_manager.resolve_limits(
+        defaults, decorated = self.limit_manager.resolve_limits(
             flask.current_app,
             endpoint,
             blueprint,
+            name,
             in_middleware,
             marked_for_limiting,
-            fallback_limits,
         )
+        limits = OrderedSet(defaults) - self.context.seen_limits
+        self.context.seen_limits.update(defaults)
+        return list(limits) + list(decorated)
 
     def __evaluate_limits(self, endpoint: str, limits: List[Limit]) -> None:
         failed_limits: List[Tuple[Limit, List[str]]] = []
         limit_for_header: Optional[RequestLimit] = None
         view_limits: List[RequestLimit] = []
         for lim in sorted(limits, key=lambda x: x.limit):
-            limit_scope = lim.scope or endpoint
-
             if lim.is_exempt or lim.method_exempt:
                 continue
 
-            if lim.per_method:
-                limit_scope += f":{flask.request.method}"
+            limit_scope = lim.scope_for(endpoint, flask.request.method)
             limit_key = lim.key_func()
             args = [limit_key, limit_scope]
             kwargs = {}
@@ -955,7 +978,7 @@ class Limiter:
             view_limits.append(RequestLimit(self, lim.limit, args, False))
 
             if not method(lim.limit, *args, **kwargs):
-                self.logger.warning(
+                self.logger.info(
                     "ratelimit %s (%s) exceeded at endpoint: %s",
                     lim.limit,
                     limit_key,
@@ -980,21 +1003,28 @@ class Limiter:
                         if isinstance(cb_response, flask.wrappers.Response):
                             on_breach_response = cb_response
                     except Exception as err:  # noqa
-                        self.logger.warning(
-                            "on_breach callback failed with error %s", err
-                        )
-
+                        if self._swallow_errors:
+                            self.logger.exception(
+                                "on_breach callback failed with error %s", err
+                            )
+                        else:
+                            raise err
         if failed_limits:
             raise RateLimitExceeded(
                 sorted(failed_limits, key=lambda x: x[0].limit)[0][0],
                 response=on_breach_response,
             )
 
-    def _check_request_limit(self, in_middleware: bool = True) -> None:
-        endpoint = flask.request.endpoint or ""
+    def _check_request_limit(
+        self, callable_name: Optional[str] = None, in_middleware: bool = True
+    ) -> None:
+        endpoint = self.identify_request()
         try:
             all_limits = self.__filter_limits(
-                flask.request.endpoint, flask.request.blueprint, in_middleware
+                endpoint,
+                flask.request.blueprint,
+                callable_name,
+                in_middleware,
             )
             self.__evaluate_limits(endpoint, all_limits)
         except Exception as e:
@@ -1007,7 +1037,10 @@ class Limiter:
                     " in-memory storage"
                 )
                 self._storage_dead = True
-                self._check_request_limit(in_middleware)
+                self.context.seen_limits.clear()
+                self._check_request_limit(
+                    callable_name=callable_name, in_middleware=in_middleware
+                )
             else:
                 if self._swallow_errors:
                     self.logger.exception("Failed to rate limit. Swallowing error")
@@ -1015,8 +1048,7 @@ class Limiter:
                     raise e
 
     def __release_context(self, _: Optional[BaseException] = None) -> None:
-        if self.context:
-            self.context.reset()
+        self.context.reset()
 
 
 class LimitDecorator:
@@ -1047,15 +1079,61 @@ class LimitDecorator:
         self.limiter: weakref.ProxyType[Limiter] = weakref.proxy(limiter)
         self.limit_value = limit_value
         self.key_func = key_func or self.limiter._key_func
-        self.scope = scope if shared else None
+        self.scope = scope
         self.per_method = per_method
-        self.methods = methods
+        self.methods = tuple(methods) if methods else None
         self.error_message = error_message
         self.exempt_when = exempt_when
         self.override_defaults = override_defaults
         self.deduct_when = deduct_when
         self.on_breach = on_breach
         self.cost = cost
+        self.is_static = not callable(self.limit_value)
+        self.shared = shared
+
+    @property
+    def limit_group(self) -> LimitGroup:
+        return LimitGroup(
+            limit_provider=self.limit_value,
+            key_function=self.key_func,
+            scope=self.scope,
+            per_method=self.per_method,
+            methods=self.methods,
+            error_message=self.error_message,
+            exempt_when=self.exempt_when,
+            override_defaults=self.override_defaults,
+            deduct_when=self.deduct_when,
+            on_breach=self.on_breach,
+            cost=self.cost,
+            shared=self.shared,
+        )
+
+    def __enter__(self) -> None:
+        tb = traceback.extract_stack(limit=2)
+        qualified_location = f"{tb[0].filename}:{tb[0].name}:{tb[0].lineno}"
+
+        # TODO: if use as a context manager becomes interesting/valuable
+        #  a less hacky approach than using the traceback and piggy backing
+        #  on the limit manager's knowledge of decorated limits might be worth it.
+        self.limiter.limit_manager.add_decorated_limit(
+            qualified_location, self.limit_group, override=True
+        )
+
+        self.limiter.limit_manager.add_endpoint_hint(
+            self.limiter.identify_request(), qualified_location
+        )
+
+        self.limiter._check_request_limit(
+            in_middleware=False, callable_name=qualified_location
+        )
+
+    def __exit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc_value: Optional[BaseException],
+        traceback: Optional[TracebackType],
+    ) -> None:
+        ...
 
     @overload
     def __call__(self, obj: Callable[P, R]) -> Callable[P, R]:
@@ -1068,81 +1146,44 @@ class LimitDecorator:
     def __call__(
         self, obj: Union[Callable[P, R], flask.Blueprint]
     ) -> Optional[Callable[P, R]]:
-        is_route = not isinstance(obj, flask.Blueprint)
         if isinstance(obj, flask.Blueprint):
             name = obj.name
         else:
-            name = f"{obj.__module__}.{obj.__name__}"
-
-        dynamic_limit, static_limits = None, []
-
-        if callable(self.limit_value):
-            dynamic_limit = LimitGroup(
-                self.limit_value,
-                self.key_func,
-                self.scope,
-                self.per_method,
-                self.methods,
-                self.error_message,
-                self.exempt_when,
-                self.override_defaults,
-                self.deduct_when,
-                self.on_breach,
-                self.cost,
-            )
-        else:
-            try:
-                static_limits = list(
-                    LimitGroup(
-                        self.limit_value,
-                        self.key_func,
-                        self.scope,
-                        self.per_method,
-                        self.methods,
-                        self.error_message,
-                        self.exempt_when,
-                        self.override_defaults,
-                        self.deduct_when,
-                        self.on_breach,
-                        self.cost,
-                    )
-                )
-            except ValueError as e:
-                self.limiter.logger.error(
-                    "failed to configure %s %s (%s)",
-                    "view function" if is_route else "blueprint",
-                    name,
-                    e,
-                )
+            name = get_qualified_name(obj)
 
         if isinstance(obj, flask.Blueprint):
-            if dynamic_limit:
-                self.limiter.limit_manager.add_runtime_blueprint_limits(
-                    name, dynamic_limit
-                )
-            else:
-                self.limiter.limit_manager.add_static_blueprint_limits(
-                    name, *static_limits
-                )
+            self.limiter.limit_manager.add_blueprint_limit(name, self.limit_group)
             return None
         else:
             self.limiter._marked_for_limiting.add(name)
-
-            if dynamic_limit:
-                self.limiter.limit_manager.add_runtime_route_limits(name, dynamic_limit)
-            else:
-                self.limiter.limit_manager.add_static_route_limits(name, *static_limits)
+            self.limiter.limit_manager.add_decorated_limit(name, self.limit_group)
 
             @wraps(obj)
             def __inner(*a: P.args, **k: P.kwargs) -> R:
                 if (
                     self.limiter._auto_check
-                    and not self.limiter.context.rate_limiting_complete
+                    and not getattr(obj, "__wrapper-limiter-instance", None)
+                    == self.limiter
                 ):
-                    self.limiter._check_request_limit(False)
-                    self.limiter.context.rate_limiting_complete = True
+                    identity = self.limiter.identify_request()
+                    if identity:
+                        view_func = flask.current_app.view_functions.get(identity, None)
+                        if view_func and not get_qualified_name(view_func) == name:
+                            self.limiter.limit_manager.add_endpoint_hint(identity, name)
+
+                    self.limiter._check_request_limit(
+                        in_middleware=False, callable_name=name
+                    )
+
                 return cast(
                     R, flask.current_app.ensure_sync(cast(Callable[P, R], obj))(*a, **k)
                 )
 
+            # mark this wrapper as wrapped by a decorator from the limiter
+            # from which the decorator was created. This ensures that stacked
+            # decorations only trigger rate limiting from the inner most
+            # decorator from each limiter instance (the weird need for
+            # keeping track of the instance is to handle cases where multiple
+            # limiter extensions are registered on the same application).
+            setattr(__inner, "__wrapper-limiter-instance", self.limiter)
             return __inner
