@@ -190,7 +190,7 @@ def test_key_func(extension_factory):
     app, limiter = extension_factory()
 
     @app.route("/t1")
-    @limiter.limit("100 per minute", lambda: "test")
+    @limiter.limit("100 per minute", key_func=lambda: "test")
     def t1():
         return "test"
 
@@ -207,8 +207,9 @@ def test_key_func(extension_factory):
 
 
 def test_logging(caplog):
+    caplog.set_level(logging.INFO)
     app = Flask(__name__)
-    limiter = Limiter(app, key_func=get_remote_address)
+    limiter = Limiter(get_remote_address, app=app)
 
     @app.route("/t1")
     @limiter.limit("1/minute")
@@ -219,15 +220,16 @@ def test_logging(caplog):
         assert 200 == cli.get("/t1").status_code
         assert 429 == cli.get("/t1").status_code
     assert len(caplog.records) == 1
-    assert caplog.records[0].levelname == "WARNING"
+    assert caplog.records[0].levelname == "INFO"
 
 
-def test_reuse_logging():
+def test_reuse_logging(caplog):
+    caplog.set_level(logging.INFO)
     app = Flask(__name__)
     app_handler = mock.Mock()
     app_handler.level = logging.INFO
     app.logger.addHandler(app_handler)
-    limiter = Limiter(app, key_func=get_remote_address)
+    limiter = Limiter(get_remote_address, app=app)
 
     for handler in app.logger.handlers:
         limiter.logger.addHandler(handler)
@@ -271,7 +273,7 @@ def test_multiple_apps():
     app1 = Flask(__name__)
     app2 = Flask(__name__)
 
-    limiter = Limiter(default_limits=["1/second"], key_func=get_remote_address)
+    limiter = Limiter(get_remote_address, default_limits=["1/second"])
     limiter.init_app(app1)
     limiter.init_app(app2)
 
@@ -322,10 +324,10 @@ def test_multiple_apps():
 def test_headers_no_breach():
     app = Flask(__name__)
     limiter = Limiter(
-        app,
+        get_remote_address,
+        app=app,
         default_limits=["10/minute"],
         headers_enabled=True,
-        key_func=get_remote_address,
     )
 
     @app.route("/t1")
@@ -355,10 +357,10 @@ def test_headers_no_breach():
 def test_headers_breach():
     app = Flask(__name__)
     limiter = Limiter(
-        app,
+        get_remote_address,
+        app=app,
         default_limits=["10/minute"],
         headers_enabled=True,
-        key_func=get_remote_address,
     )
 
     @app.route("/t1")
@@ -384,10 +386,10 @@ def test_headers_breach():
 def test_retry_after():
     app = Flask(__name__)
     _ = Limiter(
-        app,
+        get_remote_address,
+        app=app,
         default_limits=["1/minute"],
         headers_enabled=True,
-        key_func=get_remote_address,
     )
 
     @app.route("/t1")
@@ -407,10 +409,10 @@ def test_retry_after():
 def test_retry_after_exists_seconds():
     app = Flask(__name__)
     _ = Limiter(
-        app,
+        get_remote_address,
+        app=app,
         default_limits=["1/minute"],
         headers_enabled=True,
-        key_func=get_remote_address,
     )
 
     @app.route("/t1")
@@ -427,10 +429,10 @@ def test_retry_after_exists_seconds():
 def test_retry_after_exists_rfc1123():
     app = Flask(__name__)
     _ = Limiter(
-        app,
+        get_remote_address,
+        app=app,
         default_limits=["1/minute"],
         headers_enabled=True,
-        key_func=get_remote_address,
     )
 
     @app.route("/t1")
@@ -450,10 +452,10 @@ def test_custom_headers_from_config():
     app.config.setdefault(ConfigVars.HEADER_REMAINING, "X-Remaining")
     app.config.setdefault(ConfigVars.HEADER_RESET, "X-Reset")
     limiter = Limiter(
-        app,
+        get_remote_address,
+        app=app,
         default_limits=["10/minute"],
         headers_enabled=True,
-        key_func=get_remote_address,
     )
 
     @app.route("/t1")
@@ -515,7 +517,6 @@ def test_callable_default_limit(extension_factory):
 
 
 def test_callable_application_limit(extension_factory):
-
     app, limiter = extension_factory(
         application_limits=[
             lambda: request.headers.get("suspect", 0) and "1/minute" or "2/minute"
@@ -686,11 +687,11 @@ def test_custom_key_prefix(redis_connection, extension_factory):
         assert 429 == resp.status_code
 
 
-def test_second_instance_bypassed_by_shared_g():
+def test_multiple_instances_no_key_prefix():
     app = Flask(__name__)
-    limiter1 = Limiter(app, key_func=get_remote_address)
+    limiter1 = Limiter(get_remote_address, app=app)
 
-    limiter2 = Limiter(app, key_func=get_remote_address)
+    limiter2 = Limiter(get_remote_address, app=app)
 
     @app.route("/test1")
     @limiter2.limit("1/second")
@@ -706,27 +707,26 @@ def test_second_instance_bypassed_by_shared_g():
     with hiro.Timeline().freeze() as timeline:
         with app.test_client() as cli:
             assert cli.get("/test1").status_code == 200
-            assert cli.get("/test2").status_code == 200
             assert cli.get("/test1").status_code == 429
             assert cli.get("/test2").status_code == 200
+            assert cli.get("/test2").status_code == 429
 
             for i in range(8):
-                assert cli.get("/test1").status_code == 429
+                timeline.forward(1)
+                assert cli.get("/test1").status_code == 200
                 assert cli.get("/test2").status_code == 200
-            assert cli.get("/test2").status_code == 429
             timeline.forward(1)
             assert cli.get("/test1").status_code == 200
             assert cli.get("/test2").status_code == 429
             timeline.forward(59)
-            assert cli.get("/test1").status_code == 200
             assert cli.get("/test2").status_code == 200
 
 
 def test_independent_instances_by_key_prefix():
     app = Flask(__name__)
-    limiter1 = Limiter(app, key_prefix="lmt1", key_func=get_remote_address)
+    limiter1 = Limiter(get_remote_address, key_prefix="lmt1", app=app)
 
-    limiter2 = Limiter(app, key_prefix="lmt2", key_func=get_remote_address)
+    limiter2 = Limiter(get_remote_address, key_prefix="lmt2", app=app)
 
     @app.route("/test1")
     @limiter2.limit("1/second")
@@ -766,14 +766,12 @@ def test_independent_instances_by_key_prefix():
 
 def test_multiple_limiters_default_limits():
     app = Flask(__name__)
+    Limiter(get_remote_address, key_prefix="lmt1", app=app, default_limits=["1/second"])
     Limiter(
-        app, key_prefix="lmt1", default_limits=["1/second"], key_func=get_remote_address
-    )
-    Limiter(
-        app,
+        get_remote_address,
         key_prefix="lmt2",
         default_limits=["10/minute"],
-        key_func=get_remote_address,
+        app=app,
     )
 
     @app.route("/test1")
